#ifndef __STRING_HXX
#define __STRING_HXX

//
//
/*  Copyright Digital Equipment Corporation 1990, 1995. All rights reserved.*/
//
/*  Restricted Rights: Use, duplication, or disclosure by the U.S.*/
/*  Government is subject to restrictions as set forth in subparagraph*/
/*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR*/
/*  52.227-14 Alt. III, as applicable.*/
//
/*  This software is proprietary to and embodies the confidential*/
/*  technology of Digital Equipment Corporation. Possession, use, or*/
/*  copying of this software and media is authorized only pursuant to a*/
/*  valid written license from Digital or an authorized sublicensor.*/
//
//

/*++*/
/*  FACILITY:	DEC C++ Class Library*/
//
/*  SUBSYSTEM:	The String package.*/
//
/*  MODULE:	Include file for the String package.*/
//
/*  MODULE DESCRIPTION:*/
//
/*	This module implements the basic character string operations.*/
//
/*  DESIGN ISSUES:*/
//
/*      The interface of this String package is based on the documentation for*/
/*      AT&T C++ Standard Library Extension Release 1.0 and the examples found*/
/*	in the following books:  "The C++ Programming Language" by Bjarne*/
/*	Stroustrup, "The C++ Answer Book" by Tony L. Hansen, and "C++ Primer"*/
/*	by Stanley B. Lippman.*/
//
/*--*/


/*++*/
/*  INCLUDE FILES*/
/*--*/

#include <cxxl.hxx>
#include <iostream.hxx>

#ifdef _NT
#include <objectio.hxx>
#else
#include <objection.hxx>
#endif

extern "C"
{
#include <string.h>
}

#ifdef __PRAGMA_ENVIRONMENT
#   pragma __environment save
#   pragma __environment header_defaults
#else
#   if defined(__VMS) && defined(__DECCXX)
#      pragma __extern_model __save
#      pragma __extern_prefix __save
#      pragma __member_alignment __save
#   endif
#endif

/*++*/
/*  LINKAGE*/
/*--*/

#if defined(__VMS) && defined(__DECCXX)
#   pragma __extern_model __relaxed_refdef
#   pragma __extern_prefix __CXXL_PLAIN_FUNC_LINKAGE
#   if defined __VAX
#       pragma __nomember_alignment
#   else
#       pragma __member_alignment
#   endif
#endif


/*++*/
/*  EXTERNAL DATA*/
/*--*/

/* extern Objection_action cxxl_String_error;*/

class String

/*++*/
/*  CLASS DESCRIPTION:*/
//
/*	A String is a sequence of characters of type char.  The sequence of*/
/*	characters is stored as a null-terminated array (ASCIZ format).*/
/*	Therefore, a String may NOT contain a null character.  More than one*/
/*	String may share the same data (string representation structure).*/
/*	Garbage collection is done with a reference count in the string*/
/*	representation.*/
/*--*/

{
private:
    struct		cxxl_srep		/* String's representation*/
			{
			    char *cxxl_s;	/* Pointer to the data*/
			    int  cxxl_n;	/* Reference count*/
			};
    cxxl_srep		*cxxl_p;	/* Pointer to String representation*/

public:
			String();
			String(const String &);
			String(const char *);
			String(const char &);
			~String();

    /* Objections*/
    /* static Objection	memory_error;*/

    /* Assignments*/
    String		&operator=(const String &);
    String		&operator=(const char *);

    /* Conversions*/
    inline		operator char *() const;
    inline		operator const char *() const;

    /* Operators*/
    friend inline ostream	&operator<<(ostream &, const String &);
    friend istream		&operator>>(istream &, String &);
    friend inline int		operator==(const String &, const String &);
    friend inline int		operator==(const String &, const char *);
    friend inline int		operator==(const char *, const String &);
    friend inline int		operator!=(const String &, const String &);
    friend inline int		operator!=(const String &, const char *);
    friend inline int		operator!=(const char *, const String &);
    friend inline int		operator<(const String &, const String &);
    friend inline int		operator<(const String &, const char *);
    friend inline int		operator<(const char *, const String &);
    friend inline int		operator>(const String &, const String &);
    friend inline int		operator>(const String &, const char *);
    friend inline int		operator>(const char *, const String &);
    friend inline int		operator<=(const String &, const String &);
    friend inline int		operator<=(const String &, const char *);
    friend inline int		operator<=(const char *, const String &);
    friend inline int		operator>=(const String &, const String &);
    friend inline int		operator>=(const String &, const char *);
    friend inline int		operator>=(const char *, const String &);

    friend String		operator+(const String &, const String &);
    friend String		operator+(const String &, const char *);
    friend String		operator+(const char *, const String &);

    String			&operator+=(const String &);
    String			&operator+=(const char *);
    String			operator()(int, int);
    String			operator()(int, int) const;
    inline char			operator[](int) const;
    inline char			&operator[](int);

    /* Additional member functions*/
    inline unsigned int	length() const;
    String		upper() const;
    String		lower() const;
    int			match(const String &) const;
    int			index(const String &) const;

/*  The following functions are obsolete and only provided for binary*/
/*  compatibility with the previous iostream implementation on DEC OSF/1*/
/*  and OpenVMS for VAX*/
#if defined(__CXXL_STRING_CXX) && defined(__osf__) || defined (__VAX)
    String		&operator+=(char *);
    String		upper();
    String		lower();
    int			match(const String &);
    int			index(const String &);
#endif

    friend String		operator+(const String &, char *);
    friend String		operator+(char *, const String &);
};

inline String::operator char *(
    ) const

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is a conversion operator from String to char *.*/
//
/*  RETURN VALUE:*/
//
/*      The char * representation of the String, as internally stored in the*/
/*	cxxl_s part of its cxxl_p private data member.*/
/*--*/

{
    return cxxl_p->cxxl_s;
}

inline String::operator const char *(
    ) const

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is a conversion operator from String to const char *.*/
//
/*  RETURN VALUE:*/
//
/*      The char * representation of the String, as internally stored in the*/
/*	cxxl_s part of its cxxl_p private data member.*/
/*--*/

{
    return (const char *)(cxxl_p->cxxl_s);
}

inline ostream &operator<<(
    ostream	 &os,  /* Reference to the ostream object.*/
    const String &str  /* Reference to the String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the output operator, inserting the contents of a String object,*/
/*	specified by the str argument, into an ostream object, specified by the*/
/*	os argument.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*      Reference to the ostream object used for output.*/
/*--*/

{
    return os << str.cxxl_p->cxxl_s;
}

inline int operator==(
    const String &str_1,  /* Reference to the first String object.*/
    const String &str_2   /* Reference to the second String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the equality operator used to compare character strings*/
/*	associated with two String objects, specified by the str_1 and str_2*/
/*	arguments.  The terms of comparison are as defined for the ANSI C*/
/*	strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*      1 if the respective character strings are equal; 0 otherwise.*/
/*--*/

{
    return strcmp(str_1.cxxl_p->cxxl_s, str_2.cxxl_p->cxxl_s) == 0;
}

inline int operator==(
    const String &str,  /* Reference to the String object.*/
    const char	 *ptr   /* Pointer to the character string.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the equality operator used to compare the character string*/
/*	associated with a String object, specified by the str argument, with a*/
/*	character string, specified by the ptr argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*      1 if the respective character strings are equal; 0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) == 0;
}

inline int operator==(
    const char   *ptr,  /* Pointer to the character string.*/
    const String &str   /* Reference to the String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the equality operator used to compare a character string,*/
/*	specified by the ptr argument, with the character string associated*/
/*	with a String object, specified by the str argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*      1 if the respective character strings are equal; 0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) == 0;
}

inline int operator!=(
    const String &str_1,  /* Reference to the first String object.*/
    const String &str_2   /* Reference to the second String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the inequality operator used to compare character strings*/
/*	associated with two String objects, specified by the arguments str_1*/
/*	and str_2, respectively.  The terms of comparison are as defined for*/
/*	the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*      1 if the respective character strings are not equal; 0 otherwise.*/
/*--*/

{
    return strcmp(str_1.cxxl_p->cxxl_s, str_2.cxxl_p->cxxl_s) != 0;
}

inline int operator!=(
    const String &str,  /* Reference to the String object.*/
    const char   *ptr   /* Pointer to the character string.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the inequality operator used to compare the character string*/
/*	associated with a String object, specified by the str argument, with a*/
/*	character string, specified by the ptr argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*      1 if the respective character strings are not equal; 0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) != 0;
}

inline int operator!=(
    const char   *ptr,  /* Pointer to the character string.*/
    const String &str   /* Reference to the String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the inequality operator used to compare a character string,*/
/*	specified by the ptr argument, with the character string associated*/
/*	with a String object, specified by the str argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*      1 if the respective character strings are not equal; 0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) != 0;
}

inline int operator<(
    const String &str_1,  /* Reference to the first String object.*/
    const String &str_2   /* Reference to the second String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the less-than operator used to compare character strings*/
/*	associated with two String objects, specified by the arguments str_1*/
/*	and str_2, respectively.  The terms of comparison are as defined for*/
/*	the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is less than the second; 0 otherwise.*/
/*--*/

{
    return strcmp(str_1.cxxl_p->cxxl_s, str_2.cxxl_p->cxxl_s) < 0;
}

inline int operator<(
    const String &str,	/* Reference to the String object.*/
    const char   *ptr	/* Pointer to the character string.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the less-than operator used to compare the character string*/
/*	associated with a String object, specified by the str argument, with a*/
/*	character string, specified by the ptr argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is less than the second; 0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) < 0;
}

inline int operator<(
    const char   *ptr,  /* Pointer to the character string.*/
    const String &str   /* Reference to the String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the less-than operator used to compare a character string,*/
/*	specified by the ptr argument, with the character string associated*/
/*	with a String object, specified by the str argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is less than the second; 0 otherwise.*/
/*--*/

{
    return strcmp(ptr, str.cxxl_p->cxxl_s) < 0;
}

inline int operator>(
    const String &str_1,  /* Reference to the first String object.*/
    const String &str_2   /* Reference to the second String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the greater-than operator used to compare character strings*/
/*	associated with two String objects, specified by the str_1 and str_2*/
/*	arguments, respectively.  The terms of comparison are as defined for*/
/*	the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is greater than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(str_1.cxxl_p->cxxl_s, str_2.cxxl_p->cxxl_s) > 0;
}

inline int operator>(
    const String &str,  /* Reference to the String object.*/
    const char   *ptr   /* Pointer to the character string.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the greater-than operator used to compare the character string*/
/*	associated with a String object, specified by the str argument, with a*/
/*	character string, specified by the ptr argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is greater than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) > 0;
}

inline int operator>(
    const char   *ptr,  /* Pointer to the character string.*/
    const String &str   /* Reference to the String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the greater-than operator used to compare a character string,*/
/*	specified by the ptr argument, with the character string associated*/
/*	with a String object, specified by the str argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is greater than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(ptr, str.cxxl_p->cxxl_s) > 0;
}

inline int operator<=(
    const String &str_1,  /* Reference to the first String object.*/
    const String &str_2   /* Reference to the second String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the not-greater-than operator used to compare character strings*/
/*	associated with two String objects, specified by the str_1 and str_2*/
/*	arguments, respectively.  The terms of comparison are as defined for*/
/*	the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is not greater than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(str_1.cxxl_p->cxxl_s, str_2.cxxl_p->cxxl_s) <= 0;
}

inline int operator<=(
    const String &str,  /* Reference to the String object.*/
    const char   *ptr   /* Pointer to the character string.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the not-greater-than operator used to compare the character*/
/*	string associated with a String object, specified by the str argument,*/
/*	with a character string, specified by the ptr argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is not greater than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) <= 0;
}

inline int operator<=(
    const char   *ptr,  /* Pointer to the character string.*/
    const String &str   /* Reference to the String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the not-greater-than operator used to compare a character*/
/*	string, specified by the ptr argument, with the character string*/
/*	associated with a String object, specified by the str argument.  The*/
/*	terms of comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is not greater than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(ptr, str.cxxl_p->cxxl_s) <= 0;
}

inline int operator>=(
    const String &str_1,  /* Reference to the first String object.*/
    const String &str_2   /* Reference to the second String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the not-less-than operator used to compare character strings*/
/*	associated with two String objects, specified by the str_1 and str_2*/
/*	arguments, respectively.  The terms of comparison are as defined for*/
/*	the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is not less than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(str_1.cxxl_p->cxxl_s, str_2.cxxl_p->cxxl_s) >= 0;
}

inline int operator>=(
    const String &str,  /* Reference to the String object.*/
    const char   *ptr   /* Pointer to the character string.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the not-less-than operator used to compare the character string*/
/*	associated with a String object, specified by the str argument, with a*/
/*	character string, specified by the ptr argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is not less than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(str.cxxl_p->cxxl_s, ptr) >= 0;
}

inline int operator>=(
    const char   *ptr,  /* Pointer to the character string.*/
    const String &str   /* Reference to the String object.*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is the not-less-than operator used to compare a character string,*/
/*	specified by the ptr argument, with the character string associated*/
/*	with a String object, specified by the str argument.  The terms of*/
/*	comparison are as defined for the ANSI C strcmp function.*/
//
/*  FRIEND OF:	String*/
//
/*  RETURN VALUE:*/
//
/*	1 if the first character string is not less than the second;*/
/*	0 otherwise.*/
/*--*/

{
    return strcmp(ptr, str.cxxl_p->cxxl_s) >= 0;
}

inline char String::operator[](
    int position 	/* Index into string*/
    )
    const

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is a subscript operator ([]) which returns a character within*/
/*	the string.*/
//
/*  RETURN VALUE:*/
//
/*	The character at the requested position within the string.*/
/*	If the index is pointing past the end of the string, 0 is*/
/*	returned.  If the index is negative, the results are undefined.*/
/*--*/

{
    return (position < length()) ? cxxl_p->cxxl_s[position] : 0;
}

inline char & String::operator[](
    int position 	/* Index into string*/
    )

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This is a subscript operator ([]) which returns the reference to a */
/*	character within the string.*/
//
/*  RETURN VALUE:*/
//
/*	The reference to the character at the requested position within the */
/*	string.  This reference is potentially invalid after any subsequent */
/*	call to a non-const member function for the object.*/
/*	*/
/*	If the index is pointing past the end of the string or if the index*/
/*	is negative, the results are undefined.*/
/*--*/
{
    return cxxl_p->cxxl_s[position];
}

inline unsigned int String::length(
    ) const

/*++*/
/*  FUNCTIONAL DESCRIPTION:*/
//
/*	This member function returns the length of the character string*/
/*	associated with the String object.*/
//
/*  RETURN VALUE:*/
//
/*      The length of the character string, as defined for the ANSI C strlen*/
/*	function.*/
/*--*/

{
    return strlen(cxxl_p->cxxl_s);
}

#ifdef __PRAGMA_ENVIRONMENT
#   pragma __environment restore
#else
#   if defined(__VMS) && defined(__DECCXX)
#      pragma __extern_model  __restore
#      pragma __extern_prefix __restore
#      pragma __member_alignment __restore
#   endif
#endif

#endif
/* End of string.hxx*/
