/*
 * *****************************************************************
 * *                                                               *
 * *    Copyright (c) Digital Equipment Corporation, 1991, 1997    *
 * *                                                               *
 * *   All Rights Reserved.  Unpublished rights  reserved  under   *
 * *   the copyright laws of the United States.                    *
 * *                                                               *
 * *   The software contained on this media  is  proprietary  to   *
 * *   and  embodies  the  confidential  technology  of  Digital   *
 * *   Equipment Corporation.  Possession, use,  duplication  or   *
 * *   dissemination of the software and media is authorized only  *
 * *   pursuant to a valid written license from Digital Equipment  *
 * *   Corporation.                                                *
 * *                                                               *
 * *   RESTRICTED RIGHTS LEGEND   Use, duplication, or disclosure  *
 * *   by the U.S. Government is subject to restrictions  as  set  *
 * *   forth in Subparagraph (c)(1)(ii)  of  DFARS  252.227-7013,  *
 * *   or  in  FAR 52.227-19, as applicable.                       *
 * *                                                               *
 * *****************************************************************
 */
/*
 * HISTORY
 */
/*
 * @(#)$RCSfile: pthread.h,v $ $Revision: 1.1.13.6 $ (DEC) $Date: 1997/04/25 18:21:57 $
 */

/*
 *  FACILITY:
 *
 *	DECthreads POSIX 1003.1c
 *
 *  ABSTRACT:
 *
 *	External definitions for POSIX 1003.1c interfaces
 *
 *	General notice: some of the types defined and used in this header are
 *	not standard POSIX 1003.1c types. The DECthreads specific types are
 *	generally used instead of "void *" and "long" where pointer size
 *	#pragma declarations are necessary to support the OpenVMS Alpha
 *	64-bit programming environment. (And in a few other cases where it's
 *	convenient.) These types should not present any obstacle to programs
 *	coded using the standard types. The "addon" types are distinguished
 *	by naming convention -- they begin with "__pthread" and the remainder
 *	appears in mixed-case without punctuation. For example,
 *	"__pthreadLongAddr_t" for "void *" and "__pthreadLongUint_t" for
 *	"unsigned long".
 *
 *	This header file contains extensive comments regarding the structures
 *	and options. Please read them carefully before relying upon any
 *	undocumented features -- they're undocumented for a reason! We want
 *	you to be able to take advantage of them where appropriate, but you
 *	must also understand the restrictions and risks.
 *
 *  AUTHORS:
 *
 *	Dave Butenhof
 *
 *  CREATION DATE:
 *
 *	January 1995
 *
 *  MODIFIED BY:
 *
 *	Dave Butenhof
 *	Paul Clark
 *	Peter Portante
 *	Webb Scales
 *	Brian Silver
 *
 *  LAST MODIFICATION DATE:
 *
 *	14 April 1997
 */

#ifndef _PTHREAD_HEADER
# define _PTHREAD_HEADER

/******************************************************************************
 *
 * NOTICE: Temporary support for POSIX 1003.4a/D4 migration
 * --------------------------------------------------------
 *
 * Support 1003.4a/D4 ("DCE threads") source migration mode by allowing
 * programmer to cause #include <pthread.h> to define the old interfaces
 * (with modified names) if the _PTHREAD_USE_D4 macro is defined. This is
 * equivalent to #include <pthread_d4.h> -- the remainder of this header
 * file is not processed.
 *
 * In practice, you can invoke the source migration mode either by the
 * following sequence within your .c or .h files:
 *
 *	#define _PTHREAD_USE_D4
 *	#include <pthread.h>
 *
 * or by modifying your Makefile (or manual build sequence) to use the -D
 * compiler (or cpp) option:
 *
 *	cc -D_PTHREAD_USE_D4 foo.c
 *
 * The POSIX 1003.4a/D4 migration support will be retired (this clause will
 * be removed from pthread.h!) in the next major release of each operating
 * system (OpenVMS VAX, OpenVMS Alpha, and Digital UNIX) that occurs at least
 * one year after the initial release of the POSIX 1003.1c interfaces in
 * Digital UNIX 4.0 and OpenVMS 7.0.
 *
 *****************************************************************************/

#if defined (_PTHREAD_USE_D4) || defined(PTHREAD_USE_D4)
# if defined (__unix__) || defined (__unix) || defined (__unix__)
#  include <dce/pthread_d4.h>
# else
#  include <pthread_d4.h>
# endif
#else

/******************************************************************************
 *
 * NOTICE: inline function performance vs. binary compatibility
 * ------------------------------------------------------------
 *
 * This version of pthread.h supports the capability of generating inlined
 * code on OpenVMS Alpha and Digital UNIX for improved performance on some
 * critical operations -- especially mutex synchronization and reading thread
 * specific data.
 *
 * There is a balance between performance and maintainability, however.
 * Inline code sequences cannot be transparently fixed by future versions of
 * DECthreads. Thus, if a bug is found in the inlined code and later fixed,
 * programs using the inlined sequences must be recompiled to acquire the
 * fix. Additionally changes may be made to inlined sequences to improve
 * reliability or performance in the future. Digital cannot guarantee that
 * the new sequences will be binary compatible with the old sequences, and
 * therefore any such changes may also require recompilation.
 *
 * By default, this version of pthread.h will generate calls to carefully
 * tuned assembly code functions within DECthreads that implement the same
 * algorithm as the inline asm() code sequences. This provides safety at the
 * expense of performance by adding the overhead of a call and some
 * unavoidable procedure prolog and epilog code. If you need the performance
 * advantage of avoiding those calls, AND ARE WILLING TO ACCEPT THE RISK THAT
 * YOU WILL NEED TO RECOMPILE LATER, define the preprocessor symbol
 * _PTHREAD_USE_INLINE before including this header file; either by compiling
 * with
 *
 *	cc -D_PTHREAD_USE_INLINE
 *
 * or by using "#define" before including this header file:
 *
 *	#define _PTHREAD_USE_INLINE
 *	#include <pthread.h>
 *
 *****************************************************************************/

/******************************************************************************
 *
 * NOTICE: DECthreads metered sychronization and static mutex initialization
 * -------------------------------------------------------------------------
 *
 * When a program using DECthreads is run with the environment variable
 * PTHREAD_CONFIG containing the option "METER=1", DECthreads will record all
 * synchronization operations. Normally, when a mutex is statically
 * initialized using PTHREAD_MUTEX_INITIALIZER that mutex is excluded from
 * metering until the first time a thread is required to block on the mutex
 * (when contention occurs). Although this may be fine, during debugging it's
 * often useful to know when "nonblocking" locks occur, as well.
 *
 * If the preprocessor symbol _PTHREAD_NOMETER_STATIC is not defined when
 * this header file is included, statically initialized mutexes will be set
 * up such that the first attempt to lock the mutex (regardless of whether
 * the thread must block) will involve a call into DECthreads. If metering is
 * enabled the operation (and all future operations on the mutex) will be
 * metered. If metering is not enabled, the mutex will be modified such that
 * nonblocking locks (and unlocks without waiters) will not require
 * additional calls into DECthreads. In other words, when metering is not
 * enabled the extra cost of the default mode (_PTHREAD_NOMETER_STATIC not
 * defined) is a single call the first time a thread uses each mutex.
 *
 *****************************************************************************/

/******************************************************************************
 *
 * NOTICE: POSIX 1003.1c support
 * -----------------------------
 *
 * The POSIX 1003.1c-1995 standard was formally approved at the June
 * 1995 meeting of the IEEE Standards Board. The correct configuration
 * test macro for strictly conforming POSIX 1003.1c-1995 applications is
 *
 *	#define _POSIX_C_SOURCE 199506L
 *
 * The Digital UNIX <standards.h> will default _POSIX_C_SOURCE to the latest
 * POSIX 1003.1 revision supported by the operating system (currently
 * 1003.1c-1995) if no other confining standard definitions are already
 * defined. In other words, unless the programmer specifies a specific
 * restriction, the programmer will have access to all standards that aren't
 * inherently incompatible. Note that compiling with some other standards
 * modes may disable POSIX 1003.1c-1995 support. For example, defining any of
 * the symbols _POSIX_SOURCE (1003.1-1990), _POSIX_C_SOURCE=199309L
 * (1003.1b-1993), _XOPEN_SOURCE or _XOPEN_SOURCE_EXTENDED (UNIX95 or earlier,
 * which require 1003.1-1990), or _ANSI_C_SOURCE (ANSI C) will disable thread
 * support. A program built with any of these options that includes
 * <pthread.h> will not compile properly, as <sys/types.h> will fail to
 * define types and symbols required by pthread.h.
 *
 * The <unistd.h> header always defines _POSIX_VERSION to the latest POSIX.1
 * revision supported by the system -- 199506L (1003.1c-1995) for Digital
 * Unix 4.0. A portable program can use this symbol (or specific POSIX
 * feature test macros such as _POSIX_THREADS, also defined in <unistd.h>) to
 * determine whether the system supports POSIX 1003.1c-1995. (However, beware
 * that implementations of early drafts of 1003.1c, such as "DCE threads",
 * which supported 1003.4a/D4, may also define _POSIX_THREADS, but will not
 * define _POSIX_VERSION to 199506L.)
 *
 *****************************************************************************/

#if defined (__cplusplus) || defined (__DECCXX)
# define _PTHREAD_ENV_CXX
#endif

#if defined (__DECCXX) || defined (__DECC) || defined (__decc)
# define _PTHREAD_ENV_DECC
#endif

#if defined (VMS) || defined (__VMS) || defined (__vms) || defined (vms)
# define _PTHREAD_ENV_VMS
#endif

#if defined (__unix__) || defined (__unix) || defined (__unix__)
# define _PTHREAD_ENV_UNIX
#endif

#if defined (_WIN32) || defined (WIN32)
# define _PTHREAD_ENV_WIN32
#endif

#if defined (__vax__) || defined (VAX) || defined (__vax) || defined (__VAX)
# define _PTHREAD_ENV_VAX
#endif

#if defined (__alpha) || defined (__ALPHA)
# define _PTHREAD_ENV_ALPHA
#endif

#if defined (_M_IX86)
# define _PTHREAD_ENV_X86
#endif

#if !defined (_PTHREAD_ENV_DECC) && defined (_PTHREAD_ENV_VAX)
# if defined(vaxc) || defined(VAXC) || defined(__vaxc) || defined(__VAXC)
#  define _PTHREAD_ENV_VAXC
# endif
#endif

#if defined (_PTHREAD_ENV_DECC) && defined (_PTHREAD_ENV_VMS)
# if __DECC_VER >= 50200000
#  define _PTHREAD_ENV_NEWDECC
#  pragma __environment __save
# else
#  pragma __extern_model __save
# endif
# pragma __extern_model __strict_refdef
# define _PTHREAD_IMPORT_	extern
#elif defined (_PTHREAD_ENV_VAXC)
# pragma nostandard
# define _PTHREAD_IMPORT_	globalref
#else
# define _PTHREAD_IMPORT_	extern
#endif

#ifdef __cplusplus
    extern "C" {
#endif

#include <stddef.h>

#if defined(_PTHREAD_ENV_UNIX) && !defined(_PTHREAD_CORE_BUILD_)
# define _PTHREAD_EXC_INCL_CLEAN
#endif
#include <pthread_exception.h>
#ifdef _PTHREAD_EXC_INCL_CLEAN
# undef _PTHREAD_EXC_INCL_CLEAN
#endif

#ifdef _PTHREAD_ENV_UNIX
# include <standards.h>
# ifndef _PTHREAD_ENV_CXX
#  include <c_asm.h>
#  if defined (__DECC)
#   define _PTHREAD_USE_ASM_		1
#  endif
# endif
# include <unistd.h>
# define _PTHREAD_USE_MANGLED_NAMES_	1
# if defined (__DECC) && !defined (_PTHREAD_LEGACY_NP)
#  define _PTHREAD_USE_PTDNAM_
# endif
# define _PTDNAM(name) __##name
# include <sched.h>
# include <timers.h>
# include <sys/types.h>
# include <sys/time.h>
#else
# include <types.h>
# if defined (_PTHREAD_ENV_DECC) && (__DECC_VER >= 50260000)
#  include <timers.h>
# else
#  ifndef _TIMESPEC_T_
#   define _TIMESPEC_T_
typedef struct timespec {
    unsigned long	tv_sec;		/* seconds */
    long		tv_nsec;	/* nanoseconds */
    } timespec_t;
#  endif
# endif
# ifndef _SCHEDPARAM_T_
#  define _SCHEDPARAM_T_
typedef struct sched_param {
    int sched_priority;
    } sched_param_t;
# endif

# ifndef sched_yield
#  define sched_yield	pthread_yield_np
# endif
#endif

#ifdef _PTHREAD_ENV_WIN32
# include <windows.h>
# include <time.h>
# define _PTHREAD_USE_MANGLED_NAMES_ 1
# define _PTDNAM(name) __##name
# define _PTHREAD_CALL_ __stdcall
#else
# define _PTHREAD_CALL_
#endif

#ifndef _PTDNAM
# define _PTDNAM(name) name
#endif

#if defined(_PTHREAD_ENV_VMS) && defined(_PTHREAD_ENV_ALPHA)
# if defined (__DECC) && (__DECC_VER >= 50000000)
#  define _PTHREAD_USE_ASM_		1
#  if __DECC_VER < 50230004
#   define _PTHREAD_SKIP_LOOPY_ASM_	1
#  endif
#  include <c_asm.h>
# endif
# define _PTHREAD_ALLOW_MIXED_PROTOS_	1
# include <ints.h>
#endif

/*
 * Enable reentrant errno.h
 */
#ifdef _PTHREAD_ENV_UNIX
# define _REENTRANT	1
#endif
#ifdef _PTHREAD_ENV_WIN32
# define _MT		1
#endif

#if !defined (_PTHREAD_ENV_UNIX)
/*
 * If we are not a UNIX system, then the field names for the POSIX standard
 * types don't have to be name space protected.
 */
# define _Pfield(_arg_)	_arg_
#endif

#ifdef _PTHREAD_ENV_VAXC
/*
 * If compiling on OpenVMS VAX under VAX C, define missing POSIX 1003.1c
 * error symbols to the values used by DEC C so they'll match what DECthreads
 * returns.
 */
# ifdef ENOSYS
#  undef ENOSYS
# endif
# define ENOSYS		76
# ifdef ENOTSUP
#  undef ENOTSUP
# endif
# define ENOTSUP	81
# ifdef EDEADLK
#  undef EDEADLK
# endif
# define EDEADLK	82
#endif

/******************************************************************************
 *
 * NOTICE: Digital UNIX and Win32 "name mangling" in initial release
 * -----------------------------------------------------------------
 *
 * On Digital UNIX, maintaining binary compatibility for call_shared code
 * using the old 1003.4a/D4 interfaces (many of which have the same names as
 * the 1003.1c functions) requires that the compatibility interfaces retain
 * their names. Since the 1003.4a/D4 and 1003.1c functions must be able to
 * coexist within a process, this inevitably leads to the conclusion that we
 * must give the 1003.1c functions "mangled" names during the 1 year
 * 1003.4a/D4 retirement period.
 *
 * To that end, we map the 1003.1c standard names into "mangled" internal
 * names. When 1003.4a/D4 is retired these macros will be removed; the
 * 1003.1c functions will then be given their proper names. Binary
 * compatibility for images compiled during the retirement period will be
 * retained by aliases for the transition names. Note that we "mangle" only
 * the names that are identical to 1003.4a/D4 names. New functions (such as
 * pthread_key_delete) and functions with changed names (such as
 * pthread_key_create, contrasted to 1003.4a/D4's pthread_keycreate) are not
 * mangled.
 *
 * The following names are "mangled":
 *
 *	pthread_attr_getguardsize_np
 *	pthread_attr_getinheritsched
 *	pthread_attr_getstacksize
 *	pthread_attr_setguardsize_np
 *	pthread_attr_setinheritsched
 *	pthread_attr_setstacksize
 *	pthread_cancel
 *	pthread_cond_broadcast
 *	pthread_cond_destroy
 *	pthread_cond_init
 *	pthread_cond_sig_preempt_int_np
 *	pthread_cond_signal
 *	pthread_cond_signal_int_np
 *	pthread_cond_timedwait
 *	pthread_cond_wait
 *	pthread_create
 *	pthread_delay_np
 *	pthread_detach
 *	pthread_equal
 *	pthread_exit
 *	pthread_get_expiration_np
 *	pthread_getspecific
 *	pthread_join
 *	pthread_lock_global_np
 *	pthread_mutex_destroy
 *	pthread_mutex_init
 *	pthread_mutex_lock
 *	pthread_mutex_trylock
 *	pthread_mutex_unlock
 *	pthread_once
 *	pthread_self
 *	pthread_setspecific
 *	pthread_testcancel
 *	pthread_unlock_global_np
 *
 * On OpenVMS this is irrelevant since name binding occurs at link time: when
 * images are activated they bind to a known offset within the image symbol
 * vector (Alpha) or transfer vector (VAX). The binary compatibility
 * interface names have been changed, but the arrangement of the transfer
 * vector and symbol vector remain the same. Thus old code linked against the
 * old shareable images will call the correct new routine.
 *
 *****************************************************************************/

/*
 * Define a symbol which client code can test to determine whether the
 * underlying threads package is DECthreads or not. This is especially
 * important to clients of the POSIX 1003.1c interface who may want to use
 * DECthreads extensions, such as the TEB, global lock and non-real-time
 * scheduling policies while maintaining portability to a "vanilla"
 * 1003.1c-1995 implementation.
 *
 * The DECthreads version is encoded in this number. The version is normally
 * in the form "Vvv.bb-lll" where "vv" represents the major version, "bb"
 * represents the baselevel (which increases continuously regardless of
 * version), and "lll" is a build level that increases in each DECthreads
 * project build within a baselevel.
 *
 * Higher numeric values are always more recent versions. So to check for
 * support of a feature that appeared in DECthreads V3.13-100, you could
 * check whether the symbol is > 212100L.
 */

#ifdef _DECTHREADS_VERSION
# undef _DECTHREADS_VERSION
#endif
#define _DECTHREADS_VERSION 313439

#ifndef _DECTHREADS_
# define _DECTHREADS_ _DECTHREADS_VERSION
#endif

/******************************************************************************
 *
 * NOTICE: OpenVMS Alpha 64-bit pointer support
 * --------------------------------------------
 *
 * On OpenVMS Alpha DECthreads generally supports both the 64 bit and 32 bit
 * programming environments. Some external data structures (such as the TEB
 * and attributes objects) cannot vary between interfaces, and are defined to
 * contain 64 bit values regardless of programming environment. Only one
 * 1003.1c routine varies between the interfaces: pthread_join, which writes
 * a "void *" to a caller-provided address. For pthread_join32 only the low
 * 32 bits of the thread return value is written. For pthread_join64 all 64
 * bits are written (this will corrupt caller data if the allocated "void *"
 * is 32 bits wide). The default definition of "pthread_join" depends upon
 * the programming environment specified by compiler switches -- in other
 * words, the value of the _INITIAL_POINTER_SIZE macro.
 *
 * The following typedefs allow fields with required sizes to be declared
 * without needing a lot of #pragma __required_pointer_size statements
 * scattered through the header. On OpenVMS VAX and Digital UNIX these types
 * always correspond unambiguously to a single type. On OpenVMS Alpha the
 * types select the appropriate size to maintain datastructure layout
 * regardless of the default pointer size.
 *
 * Per the convention the 64-bit task force has established, a compilation on
 * OpenVMS Alpha that occurs without any /pointer_size switch should yield
 * only a 32-bit interface. But, if compilation on OpenVMS Alpha occurs with
 * an explicit /pointer_size switch, then (a) all prototypes should accept
 * long [64] bit pointers as arguments and (b) all functions which return
 * pointer values should return [32] bit pointers unless the function is a
 * variant with _64 in the name.
 *
 * The DECthreads team has no intention of supporting short pointer
 * compilation on Digital UNIX.
 *
 *	NOTE: _PTHREAD_ALLOW_MIXED_PROTOS_ should be defined only for the
 *	AXP/VMS platform.
 *
 *****************************************************************************/

#ifdef _PTHREAD_ALLOW_MIXED_PROTOS_
# ifdef __INITIAL_POINTER_SIZE
    /*
     * This is a DECC compiler with 64-bit pointer support.
     */
#  if defined(_PTHREAD_CORE_BUILD_)
#   define _PTHREAD_MIXED_PROTOS_
#  else
#   if __INITIAL_POINTER_SIZE > 0
    /*
     * The compilation has an explicit /pointer_size.
     */
#    define _PTHREAD_MIXED_PROTOS_
#   endif
#  endif
# else
    /*
     * This is some compiler on AXP/VMS that does not have 64-bit pointer
     * support.
     */
#  if defined(_PTHREAD_CORE_BUILD_)
#   error OpenVMS Alpha compiler without 64-bit pointer support
#  endif
# endif
#endif

#ifdef _PTHREAD_ALLOW_MIXED_PROTOS_
/*
 * On OpenVMS Alpha, some of the fields in DECthreads structures (TEB,
 * pthread_mutex_t, pthread_cond_t) need to be 64-bit pointers regardless of
 * the default pointer size in use by the compiler. They need to be 64 bit
 * fields even on older versions of DEC C (prior to V5.0) which don't support
 * 64 bit pointers at all.
 *
 * If the compiler supports long pointers, we use the __required_pointer_size
 * pragma to force the proper interpretation of these pointers. On older
 * compilers, we use uint64 instead.
 */
typedef int64			__pthreadLongInt_t;
typedef uint64			__pthreadLongUint_t;
# ifdef __INITIAL_POINTER_SIZE
#  pragma __required_pointer_size __save
#  pragma __required_pointer_size __long
typedef void*			__pthreadLongAddr_t;
typedef char*			__pthreadLongString_t;
typedef __pthreadLongAddr_t*	__pthreadLongAddr_p;
#  pragma __required_pointer_size __short
typedef void*			__pthreadShortAddr_t;
typedef char*			__pthreadShortString_t;
typedef __pthreadLongAddr_t*	__pthreadShortAddr_p;
#  pragma __required_pointer_size __restore
# else			/* OpenVMS Alpha with no long pointer support */
typedef uint64			__pthreadLongAddr_t;
typedef uint64			__pthreadLongString_t;
typedef uint64			__pthreadLongAddr_p;
typedef void*			__pthreadShortAddr_t;
typedef char*			__pthreadShortString_t;
typedef __pthreadLongAddr_t*	__pthreadShortAddr_p;
# endif					/* Long pointer support */
#else					/* No mixed pointers */
typedef long			__pthreadLongInt_t;
typedef unsigned long		__pthreadLongUint_t;
typedef void*			__pthreadLongAddr_t;
typedef char*			__pthreadLongString_t;
typedef __pthreadLongAddr_t*	__pthreadLongAddr_p;
typedef void*			__pthreadShortAddr_t;
typedef char*			__pthreadShortString_t;
typedef __pthreadLongAddr_t*	__pthreadShortAddr_p;
#endif

typedef void*			(*__pthreadStart_t)(void *);
typedef void			(*__pthreadDestructor_t)(void *);
typedef __pthreadLongAddr_p	__pthreadTsd_t;
typedef void			(*__pthreadOnce_t)(void);

#ifndef PTHREAD_KEYS_MAX
# define PTHREAD_KEYS_MAX		255
#endif
#ifndef PTHREAD_DESTRUCTOR_ITERATIONS
# define PTHREAD_DESTRUCTOR_ITERATIONS	4
#endif
#ifndef PTHREAD_STACK_MIN
# ifdef _PTHREAD_ENV_VAX
#  define PTHREAD_STACK_MIN		1024
# else
#  define PTHREAD_STACK_MIN		8192
# endif
#endif

#define PTHREAD_CANCEL_DISABLE		0
#define PTHREAD_CANCEL_ENABLE		1

#define PTHREAD_CANCEL_DEFERRED		0
#define PTHREAD_CANCEL_ASYNCHRONOUS	1

#define PTHREAD_CANCELED		(void *)-1

#define PTHREAD_SCOPE_PROCESS		0
#define PTHREAD_SCOPE_SYSTEM		1

#define PTHREAD_MUTEX_NORMAL_NP		0
#define PTHREAD_MUTEX_RECURSIVE_NP	1
#define PTHREAD_MUTEX_ERRORCHECK_NP	2

#define _PTHREAD_MSTATE_REFCNT	0x0000ffff	/* Internal reference count */
#define _PTHREAD_MSTATE_OVFCNT	0x000f0000	/* Overflow if not zero! */
#define _PTHREAD_MSTATE_LOCK	0x00100000	/* Primary lock bit */
#define _PTHREAD_MSTATE_LOCK_B	20		/*  (bit offset of lock) */
#define _PTHREAD_MSTATE_SLOW	0x00200000	/* Slow lock/unlock */
#define _PTHREAD_MSTATE_SLOW_B	21		/*  (bit offset of slow bit) */
#define _PTHREAD_MSTATE_TYPE	0x0f000000	/* Mutex type */
#define _PTHREAD_MSTATE_TYPE_B	24		/* (bit offset of mutex type */
#define _PTHREAD_MTYPE_NORMAL	0x00000000	/*  Normal (fast or metered) */
#define _PTHREAD_MTYPE_RECURS	0x01000000	/*  Recursive: relock in same thread */
#define _PTHREAD_MTYPE_ERRCHK	0x02000000	/*  Detect usage errors */

#define _PTHREAD_CSTATE_REFCNT	0x0000ffff	/* Internal reference count */
#define _PTHREAD_CSTATE_OVFCNT	0x000f0000	/* Overflow if not zero! */
#define _PTHREAD_CSTATE_EVENT	0x00100000	/* Might be waiters */
#define _PTHREAD_CSTATE_EVENT_B	20		/*  (bit offset of event) */
#define _PTHREAD_CSTATE_SLOW	0x00200000	/* Slow signal/broadcast */

/*
 * Validation of caller-allocated structures is minimal, but to allow basic
 * consistency checking each structure comes with a "validation" field that
 * helps to determine whether it's likely to be the correct type. Following
 * are the constants. Note that with the exception of the mutex and condition
 * variable, these constants are stored into memory passed into DECthreads,
 * so we're trusting what the caller says. Mutexes and condition variables,
 * if statically initialized using PTHREAD_COND_INITIALIZER,
 * PTHREAD_MUTEX_INITIALIZER, or their relatives, are loaded with the
 * validation constants at compile time.
 */
#define _PTHREAD_MVALID		(0x0dbcafe1L)	/* Mutex */
#define _PTHREAD_CVALID		(0x0efacbd1L)	/* Condition variable */
#define _PTHREAD_CAVALID	(0xaceface1L)	/* Condition attributes */
#define _PTHREAD_MAVALID	(0xecafeca1L)	/* Mutex attributes */
#define _PTHREAD_TAVALID	(0xbeefeed1L)	/* Thread attributes */

/*
 * The top 4 bits of the mutex valid field are reserved for DECthreads flags.
 * Setting these bits improperly may either prevent the program from running
 * or reduce the usefulness of debugging information.
 */
#define _PTHREAD_MVF_INT	0x10000000L
#define _PTHREAD_MVF_PRE	0x20000000L
#define _PTHREAD_MVF_INL	0x40000000L

#ifndef _PTHREAD_NOMETER_STATIC
# define PTHREAD_MUTEX_INITIALIZER {_PTHREAD_MSTATE_SLOW, _PTHREAD_MVALID}
# define PTHREAD_COND_INITIALIZER  {_PTHREAD_CSTATE_SLOW, _PTHREAD_CVALID}
# define PTHREAD_MUTEX_INITWITHNAME_NP(_n_,_a_) \
    {_PTHREAD_MSTATE_SLOW, _PTHREAD_MVALID, _n_, _a_}
# define PTHREAD_COND_INITWITHNAME_NP(_n_,_a_) \
    {_PTHREAD_CSTATE_SLOW, _PTHREAD_CVALID, _n_, _a_}
#else
# define PTHREAD_MUTEX_INITIALIZER {0, _PTHREAD_MVALID}
# define PTHREAD_COND_INITIALIZER  {0, _PTHREAD_CVALID}
# define PTHREAD_MUTEX_INITWITHNAME_NP(_n_,_a_) \
    {0, _PTHREAD_MVALID, _n_, _a_}
# define PTHREAD_COND_INITWITHNAME_NP(_n_,_a_) \
    {0, _PTHREAD_CVALID, _n_, _a_}
#endif
#define PTHREAD_MUTEX_INITFAST_NP {0, _PTHREAD_MVALID}
#define PTHREAD_MUTEX_INITFASTWITHNAME_NP(_n_,_a_) {0, _PTHREAD_MVALID, _n_, _a_}
#define PTHREAD_MUTEX_INITRECURSIVE_NP \
    {_PTHREAD_MTYPE_RECURS | _PTHREAD_MSTATE_SLOW, _PTHREAD_MVALID}
#define PTHREAD_MUTEX_INITRECURSIVEWITHNAME_NP(_n_,_a_) \
    {_PTHREAD_MTYPE_RECURS | _PTHREAD_MSTATE_SLOW, _PTHREAD_MVALID, _n_, _a_}
#define PTHREAD_MUTEX_INITERRCHK_NP \
    {_PTHREAD_MTYPE_ERRCHK | _PTHREAD_MSTATE_SLOW, _PTHREAD_MVALID}
#define PTHREAD_MUTEX_INITERRCHKWITHNAME_NP(_n_,_a_) \
    {_PTHREAD_MTYPE_ERRCHK | _PTHREAD_MSTATE_SLOW, _PTHREAD_MVALID, _n_, _a_}

/*
 * This macro is a special variant of the PTHREAD_MUTEX_INITIALIZER macro.
 *
 * It provides a solution for any library (or main program) that needs to
 * preempt the malloc() family of functions from libc during link. DECthreads
 * uses malloc to dynamically create an "extension" to mutex
 * (pthread_mutex_t) when a thread needs to block on the mutex, or when
 * synchronization metering is enabled. Since a thread-safe malloc will
 * generally use a mutex, there's a potential recursion in the process that
 * needs to be resolved. It can be avoided for libc, since DECthreads knows
 * which mutexes are "special". For any other malloc package that might
 * satisfy the DECthreads references to malloc, free, and valloc, another
 * mechanism is required.
 *
 * Until DECthreads initializes, use of TIS mutexes is "safe" -- they can
 * never block, and will not need the "extension". When DECthreads
 * initializes, it alters the TIS functions to call into DECthreads. If
 * metering is enabled the next lock of your malloc will need to associate an
 * extension with the mutexes involved in malloc.
 *
 * During initialization, and before making TIS "multithreaded", DECthreads
 * also preallocates a number of "mutex extension" structures and caches
 * them. These will be claimed by the first few mutexes that need to meter or
 * block. Unless metering is enabled, however, it's difficult to ensure that
 * your malloc mutexes will need an extension early enough to get one of
 * those preallocated structures.
 *
 * The solution is to initialize your malloc mutexes with this special macro.
 * The "_PTHREAD_MSTATE_SLOW" flag will ensure that the first attempt to lock
 * the mutex will call into DECthreads. This call must be made after
 * DECthreads has initialized (the TIS "stub" code won't do anything for
 * you). The special "valid" field value will cause DECthreads to immediately
 * assign that mutex an extension, even if metering isn't enabled and there's
 * no need to block. Additionally, the extension will be specially marked so
 * that DECthreads will never reassign that extension (which it might
 * otherwise do when no threads are blocked). In short, it makes the malloc
 * mutex "safe".
 *
 * The special valid field has no effect if no preallocated extension
 * structures are available, which could result in a disastrous infinite
 * recursion. That would probably mean that your code, or some other code
 * that ran between DECthreads initialization and your code, has acquired all
 * of the preallocated structures. The way around that is to reconfigure
 * DECthreads to allocate "more of them". You can do that by setting the
 * PTHREAD_CONFIG environment variable to a string including the phrase
 * "mutexes=<n>" where <n> is a number. For example,
 *
 *	export PTHREAD_CONFIG="mutexes=200"
 *
 * How big a number? That depends on how many mutexes get in your way. Try
 * calling pthread_debug_cmd("mutex -a") at the point where you want to lock
 * your malloc mutex. You should preallocate more extension structures than
 * the number of mutexes currently listed.
 */
#define _PTHREAD_MUTEX_INITPRE \
    {_PTHREAD_MSTATE_SLOW, (_PTHREAD_MVALID | _PTHREAD_MVF_PRE)}
#define _PTHREAD_MUTEX_INITPREWITHNAME(_n_,_a_) \
    {_PTHREAD_MSTATE_SLOW, (_PTHREAD_MVALID | _PTHREAD_MVF_PRE), _n_, _a_}
#define _PTHREAD_MUTEX_INITPRERECUR \
    {_PTHREAD_MTYPE_RECURS | _PTHREAD_MSTATE_SLOW, \
    (_PTHREAD_MVALID | _PTHREAD_MVF_PRE)}
#define _PTHREAD_MUTEX_INITPRERECURWITHNAME(_n_,_a_) \
    {_PTHREAD_MTYPE_RECURS | _PTHREAD_MSTATE_SLOW, \
    (_PTHREAD_MVALID | _PTHREAD_MVF_PRE), _n_, _a_}
#define _PTHREAD_MUTEX_INITPREERRCHK \
    {_PTHREAD_MTYPE_ERRCHK | _PTHREAD_MSTATE_SLOW, \
    (_PTHREAD_MVALID | _PTHREAD_MVF_PRE)}
#define _PTHREAD_MUTEX_INITPREERRCHKWITHNAME(_n_,_a_) \
    {_PTHREAD_MTYPE_ERRCHK | _PTHREAD_MSTATE_SLOW, \
    (_PTHREAD_MVALID | _PTHREAD_MVF_PRE), _n_, _a_}

#ifndef _PTHREAD_ENV_UNIX
/*
 * All references to "individual attributes" fields within these attributes
 * object structures MUST be made through the defined function calls in this
 * header file. Not all fields in these structures mean what you might infer
 * from the field type. Not all of the fields are implemented at this time.
 * The name and order of fields may change without notice, although the size
 * of the structure will never change.
 */
typedef struct __pthread_attr_t {
    long			_Pfield(valid);
    __pthreadLongString_t	_Pfield(name);
    __pthreadLongUint_t		_Pfield(arg);
    __pthreadLongUint_t		_Pfield(reserved)[19];
    } pthread_attr_t;
#endif

#ifndef _PTHREAD_ENV_UNIX
/*
 * All references to "individual attributes" fields within these attributes
 * object structures MUST be made through the defined function calls in this
 * header file. Not all fields in these structures mean what you might infer
 * from the field type. Not all of the fields are implemented at this time.
 * The name and order of fields may change without notice, although the size
 * of the structure will never change.
 */
typedef struct __pthread_mutexattr_t {
    long		_Pfield(valid);
    __pthreadLongUint_t	_Pfield(reserved)[15];
    } pthread_mutexattr_t;
#endif

#ifndef _PTHREAD_ENV_UNIX
/*
 * All references to "individual attributes" fields within these attributes
 * object structures MUST be made through the defined function calls in this
 * header file. Not all fields in these structures mean what you might infer
 * from the field type. Not all of the fields are implemented at this time.
 * The name and order of fields may change without notice, although the size
 * of the structure will never change.
 */
typedef struct __pthread_condattr_t {
    long		_Pfield(valid);
    __pthreadLongUint_t	_Pfield(reserved)[13];
    } pthread_condattr_t;
#endif

#ifndef _PTHREAD_ENV_UNIX
/*
 * DECthreads maintains a substantial amount of internal information about
 * each thread in a "Thread Control Block" that is not directly accessible to
 * the programmer. However, some of this information is required by
 * application, library, or debug/profile code in performance critical code
 * paths. Thus, DECthreads provides an "architected Thread Environment Block"
 * that makes a subset of the TCB information available directly. On Alpha
 * hardware systems the address of the TEB is available by using the CALL_PAL
 * RDUNIQ instruction. DECthreads also provides both a macro and a callable
 * function to return the TEB address, for convenience and portability.
 */
# if defined (_PTHREAD_ALLOW_MIXED_PROTOS_) && defined (__INITIAL_POINTER_SIZE)
#  pragma __required_pointer_size __save
#  pragma __required_pointer_size __long
# endif
typedef struct __pthreadTeb_t {
    __pthreadLongAddr_p	_Pfield(reserved1);	/* Reserved to DECthreads */
    __pthreadLongAddr_p	_Pfield(reserved2);	/* Reserved to DECthreads */
    unsigned short	_Pfield(size);		/* Size of TEB data structure */
    unsigned char	_Pfield(reserved3)[2];	/* Reserved to DECthreads */
    unsigned char	_Pfield(reserved4)[4];	/* Reserved to DECthreads */
    __pthreadLongUint_t	_Pfield(sequence);	/* Thread sequence number */
    __pthreadLongUint_t	_Pfield(reserved5)[2];	/* Reserved to DECthreads */
    __pthreadLongAddr_t	_Pfield(per_kt_area);	/* Pointer to kernel context */
    __pthreadLongAddr_t	_Pfield(stack_base);	/* Current stack base */
    __pthreadLongAddr_t	_Pfield(stack_reserve); /* Current stack reserve zone */
    __pthreadLongAddr_t	_Pfield(stack_yellow);	/* Current stack yellow zone */
    __pthreadLongAddr_t	_Pfield(stack_guard);	/* Current stack guard zone */
    __pthreadLongUint_t	_Pfield(stack_size);	/* Size of stack */
    __pthreadTsd_t	_Pfield(tsd_values);	/* TSD array (indexed by key) */
    unsigned long	_Pfield(tsd_count);	/* Number of TSD cells */
    unsigned int	_Pfield(reserved6);	/* Reserved to DECthreads */
    unsigned int	_Pfield(reserved7);	/* Reserved to DECthreads */
    unsigned int	_Pfield(thread_flags);	/* Dynamic external state */
    } pthreadTeb_t, *pthreadTeb_p;
# if defined (_PTHREAD_ALLOW_MIXED_PROTOS_) && defined (__INITIAL_POINTER_SIZE)
typedef pthreadTeb_p	pthread_t;	/* Long pointer if possible */
#  pragma __required_pointer_size __restore
# elif defined (_PTHREAD_ENV_ALPHA) && defined (_PTHREAD_ENV_VMS)
typedef uint64		pthread_t;	/* Force 64 bits anyway */
# else
typedef pthreadTeb_p	pthread_t;	/* Pointers is pointers */
# endif
#endif

/*
 * Macros to utilize the TEB directly for maximum efficiency. With hardware
 * and a compiler that supports it, this uses inline instructions to get
 * access to the thread's TEB... otherwise this will result in a simple call
 * into DECthreads.
 *
 * It is perfectly reasonable for code to freely access any defined TEB field
 * on any DECthreads platform, as long as you meet all of the following rules
 * (read this section carefully!)
 *
 * You must not depend in any way upon the content or meaning of any
 * "reserved" fields (reserved1, reserved2, reserved3, etc.). These are for
 * DECthreads internal use only and are subject to change without notice.
 *
 * Don't make any assumptions about the nature of the sequence number
 * returned by pthread_getsequence_np or pthread_getselfseq_np, except that
 * it is a unique integer among all threads created within the current image.
 * Consecutively created threads may not necessarily have consecutive
 * sequence numbers, nor are all possible sequence numbers necessarily
 * assigned by DECthreads. A thread created "later" does not necessarily have
 * a sequence number higher than that of a thread created "earlier".
 *
 * When coding direct access to thread specific data ("tsd") it's critical to
 * keep in mind that DECthreads optimizes a thread's tsd storage
 * requirements. Several key values can be stored in the internal thread
 * control structure; when additional keys become active DECthreads allocates
 * an extended tsd_values array, copies existing values from the control
 * structure into the extended array, and updates the TEB's tsd_values
 * pointer to point to the extended array. When a new thread is created
 * DECthreads will allocate an extended tsd array immediately if more keys
 * have been created than can fit within the control structure's minimal tsd
 * array. Thus:
 *
 *	1.	If a tsd key <tsd1> is created prior to thread <thread1>,
 *		then code running in <thread1> is guaranteed to have a
 *		pre-allocated tsd_values cell for <tsd1>.
 *	2.	If <tsd1> is created after <thread1>, then <thread1> may not
 *		have a tsd_values cell for <tsd1> until a value for that cell
 *		has been set using a DECthreads call (cma_key_set_context or
 *		pthread_setspecific).
 *
 * In case 2, it is erroneous to attempt to access the value of <tsd1> in
 * <thread1> without appropriate logic to verify the allocated size of the
 * tsd_values array (using the tsd_count field). For direct access that
 * follows this rule, see the inline version of pthread_getspecific, below.
 * The macros pthread_unchecked_getspecific_np and
 * pthread_unchecked_setspecific_np implement the fastest possible inlined
 * access to thread specific data; it is erroneous to use these unless the
 * thread is known to have an allocated cell for the tsd key.
 *
 * Only DECthreads and the O/S may use the value of the "per_kt_area"
 * pointer; for other code this is unpredictable. Modification of any TEB
 * field (other than thread specific data values within the tsd_values array
 * for which the caller owns a tsd key) may result in unpredictable behavior
 * including process termination.
 */
#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_self __pthread_self
# endif
#endif
extern _PTHREAD_CALL_ pthread_t pthread_self (void);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_ASM_
# ifdef pthread_self
#  undef pthread_self
# endif
# define pthread_self() ((pthread_t)asm ("call_pal 0x9e"))
#endif

#ifdef _PTHREAD_ENV_UNIX
# define pthread_unchecked_getspecific_np(_key_) \
    (pthread_self ()->_Pfield(tsd_values)[_key_])
# define pthread_unchecked_setspecific_np(_key_,_value_) \
    (pthread_self ()->_Pfield(tsd_values)[_key_] = (_value_))
#else
# define pthreadTeb()	((pthreadTeb_p)pthread_self())
# define pthread_unchecked_getspecific_np(_key_) \
    (pthreadTeb ()->_Pfield(tsd_values)[_key_])
# define pthread_unchecked_setspecific_np(_key_,_value_) \
    (pthreadTeb ()->_Pfield(tsd_values)[_key_] = (_value_))
#endif

#ifndef _PTHREAD_ENV_UNIX
/*
 * This is the primary structure controlling the identity and state of a
 * mutex. The "lock" field contains the primary mutex lock bit, a flag
 * indicating whether lock/unlock protocols are "fast" or "slow", the full
 * type of the mutex, and an internal reference count. The reference count
 * assures atomicity of unlock/wake sequences with respect to deletion.
 *
 * The "sequence" field gives each mutex a unique identifier, which is
 * primarily used for debugging. A new static mutex has a sequence number of
 * zero; DECthreads assigns a real sequence number when a thread first blocks
 * on the mutex. (A mutex created by calling pthread_mutex_init has a
 * sequence number assigned at creation.)
 *
 * Beware that the name and order of these fields are undocumented and
 * unsupported. Directly accessing any in your program may result in serious
 * difficulties in future releases of the operating system.
 *
 * These structures are volatile, because they contain critical shared data
 * that can't be cached by any thread. In general, the fact that operations
 * on these structures are (or "appear to be") associated with external calls
 * into DECthreads will be sufficient -- however adding volatile won't hurt
 * anything, and in any situation where volatile causes different code to be
 * generated the result should be "safer".
 */
typedef volatile struct __pthread_mutex_t {
    unsigned int	_Pfield(lock);		/* LOCK, SLOW, TYPE, REFCNT */
    unsigned int	_Pfield(valid);		/* Validation info */
    __pthreadLongString_t	_Pfield(name);	/* Name of mutex */
    unsigned int	_Pfield(arg);		/* printf argument for name */
    unsigned int	_Pfield(depth);		/* Recursive lock depth */
    unsigned long	_Pfield(sequence);	/* Mutex sequence number */
    unsigned long	_Pfield(owner);		/* Current owner (if known) */
    __pthreadLongAddr_p	_Pfield(block);		/* Pointer to blocking struct */
    } pthread_mutex_t;
#endif

#ifndef _PTHREAD_ENV_UNIX
/*
 * Similar to pthread_mutex_t, but for a condition variable.
 */
typedef volatile struct __pthread_cond_t {
    unsigned int	_Pfield(state);		/* EVENT, SLOW, REFCNT */
    unsigned int	_Pfield(valid);		/* Validation info */
    __pthreadLongString_t	_Pfield(name);	/* Name of condition variable */
    unsigned int	_Pfield(arg);		/* printf argument for name */
    unsigned long	_Pfield(sequence);	/* Condition variable seq # */
    __pthreadLongAddr_p	_Pfield(block);		/* Pointer to blocking struct */
    } pthread_cond_t;
#endif

#ifndef _PTHREAD_ENV_UNIX
typedef unsigned int	pthread_key_t;
#endif

/*
 * Ensure that an initialization routine is called exactly once.
 */
#define PTHREAD_ONCE_UNINIT	0
#define PTHREAD_ONCE_INITING	1
#define PTHREAD_ONCE_INITED	2
#define PTHREAD_ONCE_INIT	{PTHREAD_ONCE_UNINIT}
#ifndef _PTHREAD_ENV_UNIX
typedef volatile struct __pthread_once_t {
    long	_Pfield(state);
    long	_Pfield(reserved)[10];
    } pthread_once_t;
#endif

#define PTHREAD_INHERIT_SCHED		0
#define PTHREAD_EXPLICIT_SCHED		1

#define PTHREAD_CREATE_JOINABLE		0
#define PTHREAD_CREATE_DETACHED		1

#ifndef _PTHREAD_ENV_UNIX
# define SCHED_FIFO	1		/* Digital UNIX sched.h defines */
# define SCHED_RR	2		/* these constants already */
# define SCHED_OTHER	3
#endif
#define SCHED_FG_NP	SCHED_OTHER	/* "Foreground" (Timeshare) */
#define SCHED_BG_NP	(SCHED_OTHER+1) /* "Background" */
#define SCHED_LFI_NP	(SCHED_OTHER+2) /* "Low FIFO" (background FIFO) */
#define SCHED_LRR_NP	(SCHED_OTHER+3) /* "Low RR" (background RR) */

#define PRI_FIFO_MIN	16
#define PRI_FIFO_MAX	31
#define PRI_RR_MIN	16
#define PRI_RR_MAX	31
#define PRI_OTHER_MIN	PRI_FG_MIN_NP
#define PRI_OTHER_MAX	PRI_FG_MAX_NP
#define PRI_FG_MIN_NP	8
#define PRI_FG_MAX_NP	15
#define PRI_BG_MIN_NP	0
#define PRI_BG_MAX_NP	7
#define PRI_LFI_MIN_NP	0
#define PRI_LFI_MAX_NP	7
#define PRI_LRR_MIN_NP	0
#define PRI_LRR_MAX_NP	7

#define PRI_MAX		PRI_FIFO_MAX
#define PRI_MIN		PRI_BG_MIN_NP

#ifdef _PTHREAD_ENV_UNIX
extern int pthread_atfork (
	void	(*prepare)(void),
	void	(*parent)(void),
	void	(*child)(void));

extern int pthread_sigmask (
	int how,
	const sigset_t *set, sigset_t *oset);

extern int pthread_kill (
	pthread_t thread, int sig);
#endif

extern int _PTHREAD_CALL_ pthread_mutexattr_init (
	pthread_mutexattr_t	*attr);

extern int _PTHREAD_CALL_ pthread_mutexattr_destroy (
	pthread_mutexattr_t	*attr);

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_mutex_init __pthread_mutex_init
# endif
#endif
extern int _PTHREAD_CALL_ pthread_mutex_init (
	pthread_mutex_t			*mutex,
	const pthread_mutexattr_t	*attr);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_mutex_destroy __pthread_mutex_destroy
# endif
#endif
extern int _PTHREAD_CALL_ pthread_mutex_destroy (
	pthread_mutex_t		*mutex);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

extern int _PTHREAD_CALL_ pthread_mutex_block (
	pthread_mutex_t		*mutex);

extern int _PTHREAD_CALL_ pthread_mutex_tryrec (
	pthread_mutex_t		*mutex);

extern int _PTHREAD_CALL_ pthread_mutex_unblock (
	pthread_mutex_t		*mutex);

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_mutex_lock __pthread_mutex_lock
# endif
#endif
extern int _PTHREAD_CALL_ pthread_mutex_lock (
	pthread_mutex_t		*mutex);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_mutex_trylock __pthread_mutex_trylock
# endif
#endif
extern int _PTHREAD_CALL_ pthread_mutex_trylock (
	pthread_mutex_t		*mutex);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_mutex_unlock __pthread_mutex_unlock
# endif
#endif
extern int _PTHREAD_CALL_ pthread_mutex_unlock (
	pthread_mutex_t		*mutex);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif


#if defined(_PTHREAD_USE_ASM_) && !defined(_PTHREAD_SKIP_LOOPY_ASM_)

/*
 * See the note near the top of this header for the risks and benefits of
 * using inlined operations! They are NOT compiled into your program by
 * default, and if you choose to enable them you will be required to
 * recompile should DECthreads ever need to change the inline instruction
 * sequences! (While we will not do this arbitrarily, we will not hesitate
 * should it become necessary to substantially improve code performance or
 * reliability, or to fix a bug!)
 */

# define pthread_mutex_lock_prim(_m_) asm (" \
	ldah	%t1,0x10(%r31) ;\
    10: ldl_l	%v0,(%a0) ;\
	bne	%v0,30f ;\
	or	%v0,%t1,%t0 ;\
	stl_c	%t0,(%a0) ;\
	beq	%t0,20f ;\
	mb	; \
	br	%r31,30f ;\
    20: br	%r31,10b ;\
    30:", &(_m_)->_Pfield(lock))

# define pthread_mutex_unlock_prim(_m_) asm ("mb; \
	ldah	%t0,0x10(%r31) ;\
    10: ldl_l	%v0,(%a0) ;\
	bic	%v0,%t0,%t1 ;\
	cmoveq	%t1,%t1,%v0 ;\
	bne	%v0, 30f ;\
	stl_c	%t1,(%a0) ;\
	beq	%t1,20f ;\
	br	%r31,30f ;\
    20: br	%r31,10b ;\
    30: ", &(_m_)->_Pfield(lock))

# define pthread_mutex_lock_inline(_m_) \
    (pthread_mutex_lock_prim(_m_) ? pthread_mutex_block (_m_) : 0)
# define pthread_mutex_trylock_inline(_m_) \
    ((_m_)->_Pfield(lock) & _PTHREAD_MSTATE_SLOW \
    ? pthread_mutex_tryrec (_m_) \
    : (pthread_mutex_lock_prim(_m_) ? EBUSY : 0))
# define pthread_mutex_unlock_inline(_m_) (pthread_mutex_unlock_prim(_m_) \
    ? pthread_mutex_unblock (_m_) : 0)

/*
 * See the note near the top of this header for the risks and benefits of
 * using inlined operations! They are NOT compiled into your program by
 * default, and if you choose to enable them you will be required to
 * recompile should DECthreads ever need to change the inline instruction
 * sequences! (While we will not do this arbitrarily, we will not hesitate
 * should it become necessary to substantially improve code performance or
 * reliability, or to fix a bug!)
 */

# ifdef _PTHREAD_USE_INLINE
#  ifdef pthread_mutex_lock
#   undef pthread_mutex_lock
#   undef pthread_mutex_trylock
#   undef pthread_mutex_unlock
#  endif
#  define pthread_mutex_lock	pthread_mutex_lock_inline
#  define pthread_mutex_trylock pthread_mutex_trylock_inline
#  define pthread_mutex_unlock	pthread_mutex_unlock_inline
# endif
#else
# define pthread_mutex_lock_inline	pthread_mutex_lock
# define pthread_mutex_trylock_inline	pthread_mutex_trylock
# define pthread_mutex_unlock_inline	pthread_mutex_unlock
#endif

extern int _PTHREAD_CALL_ pthread_condattr_init (
	pthread_condattr_t	*attr);

extern int _PTHREAD_CALL_ pthread_condattr_destroy (
	pthread_condattr_t	*attr);

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_init __pthread_cond_init
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_init (
	pthread_cond_t			*cond,
	const pthread_condattr_t	*attr);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_destroy __pthread_cond_destroy
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_destroy (
	pthread_cond_t		*cond);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_signal __pthread_cond_signal
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_signal	(
	pthread_cond_t		*cond);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_broadcast __pthread_cond_broadcast
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_broadcast (
	pthread_cond_t		*cond);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_wait __pthread_cond_wait
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_wait (
	pthread_cond_t		*cond,
	pthread_mutex_t		*mutex);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_timedwait __pthread_cond_timedwait
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_timedwait (
	pthread_cond_t		*cond,
	pthread_mutex_t		*mutex,
	const struct timespec	*abstime);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

extern int _PTHREAD_CALL_ pthread_attr_setscope (
	pthread_attr_t		*attr,
	int			scope);

extern int _PTHREAD_CALL_ pthread_attr_getscope (
	const pthread_attr_t	*attr,
	int			*scope);

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_attr_setinheritsched __pthread_attr_setinheritsched
# endif
#endif
extern int _PTHREAD_CALL_ pthread_attr_setinheritsched (
	pthread_attr_t		*attr,
	int			inherit);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_attr_getinheritsched __pthread_attr_getinheritsched
# endif
#endif
extern int _PTHREAD_CALL_ pthread_attr_getinheritsched (
	const pthread_attr_t	*attr,
	int			*inherit);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

extern int _PTHREAD_CALL_ pthread_attr_setschedpolicy	(
	pthread_attr_t		*attr,
	int			policy);

extern int _PTHREAD_CALL_ pthread_attr_getschedpolicy	(
	const pthread_attr_t	*attr,
	int			*policy);

extern int _PTHREAD_CALL_ pthread_attr_setschedparam (
	pthread_attr_t			*attr,
	const struct sched_param	*param);

extern int _PTHREAD_CALL_ pthread_attr_getschedparam (
	const pthread_attr_t	*attr,
	struct sched_param	*param);

extern int _PTHREAD_CALL_ pthread_getschedparam (
	pthread_t		thread,
	int			*policy,
	struct sched_param	*param);

extern int _PTHREAD_CALL_ pthread_setschedparam (
	pthread_t			thread,
	int				policy,
	const struct sched_param	*param);

extern int _PTHREAD_CALL_ pthread_attr_init (
	pthread_attr_t		*attr);

extern int _PTHREAD_CALL_ pthread_attr_destroy (
	pthread_attr_t		*attr);

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_attr_setstacksize __pthread_attr_setstacksize
# endif
#endif
extern int _PTHREAD_CALL_ pthread_attr_setstacksize (
	pthread_attr_t		*attr,
	size_t			stacksize);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_attr_getstacksize __pthread_attr_getstacksize
# endif
#endif
extern int _PTHREAD_CALL_ pthread_attr_getstacksize (
	const pthread_attr_t	*attr,
	size_t			*stacksize);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

extern int _PTHREAD_CALL_ pthread_attr_setdetachstate	(
	pthread_attr_t		*attr,
	int			detachstate);

extern int _PTHREAD_CALL_ pthread_attr_getdetachstate	(
	const pthread_attr_t	*attr,
	int			*detachstate);

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_create __pthread_create
# endif
#endif
extern int _PTHREAD_CALL_ pthread_create (
	pthread_t		*thread,
	const pthread_attr_t	*attr,
	void			*(*start_routine)(void *),
	void			*arg);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_MIXED_PROTOS_
extern int _PTHREAD_CALL_ pthread_join64 (
	pthread_t		thread,
	__pthreadLongAddr_t	*value_ptr);
extern int _PTHREAD_CALL_ pthread_join32 (
	pthread_t		thread,
	__pthreadShortAddr_t	*value_ptr);
# ifdef pthread_join
#  undef pthread_join
# endif
# if __INITIAL_POINTER_SIZE == 64
#  define pthread_join pthread_join64
# else
#  define pthread_join pthread_join32
# endif
#else
# ifdef _PTHREAD_ALLOW_MIXED_PROTOS_
#  ifdef pthread_join
#   undef pthread_join
#  endif
#  define pthread_join pthread_join32
# else
#  ifdef _PTHREAD_USE_MANGLED_NAMES_
#   ifdef _PTHREAD_USE_PTDNAM_
#    pragma extern_prefix "__"
#   else
#    define pthread_join __pthread_join
#   endif
#  endif
# endif
extern int _PTHREAD_CALL_ pthread_join (
	pthread_t	thread,
	void		**value_ptr);
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix ""
# endif
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_detach __pthread_detach
# endif
#endif
extern int _PTHREAD_CALL_ pthread_detach (
	pthread_t		thread);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_exit __pthread_exit
# endif
#endif
extern void _PTHREAD_CALL_ pthread_exit (
	void			*value);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_equal __pthread_equal
# endif
#endif
extern int _PTHREAD_CALL_ pthread_equal (
	pthread_t		t1,
	pthread_t		t2);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

/*
 * See the note near the top of this header for the risks and benefits of
 * using inlined operations! They are NOT compiled into your program by
 * default, and if you choose to enable them you will be required to
 * recompile should DECthreads ever need to change the inline instruction
 * sequences! (While we will not do this arbitrarily, we will not hesitate
 * should it become necessary to substantially improve code performance or
 * reliability, or to fix a bug!)
 */

#ifdef _PTHREAD_USE_INLINE
# ifdef pthread_equal
#  undef pthread_equal
# endif
# define pthread_equal(t1,t2) (t1==t2)
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_once __pthread_once
# endif
#endif
extern int _PTHREAD_CALL_ pthread_once (
	pthread_once_t	*once_block,
	__pthreadOnce_t	init_routine);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

extern int _PTHREAD_CALL_ pthread_key_create (
	pthread_key_t		*key,
	void			(*destructor)(void *));

extern int _PTHREAD_CALL_ pthread_key_delete (
	pthread_key_t		key);

/*
 * See the note near the top of this header for the risks and benefits of
 * using inlined operations! They are NOT compiled into your program by
 * default, and if you choose to enable them you will be required to
 * recompile should DECthreads ever need to change the inline instruction
 * sequences! (While we will not do this arbitrarily, we will not hesitate
 * should it become necessary to substantially improve code performance or
 * reliability, or to fix a bug!)
 *
 * Additional note:
 *
 * pthread_getspecific can't be implemented as a macro in ANSI C, since it
 * involves two invocations of pthread_self(), which the compiler can't be
 * depended upon to combine. The next best alternative is a static inlined
 * routine; even if the #pragma inline isn't recognized by your compiler it's
 * no worse than the real call (and may be better because the compiler can
 * use optimized local call sequences); and the routine is small. Consider
 * this an experiment -- it can be used for a lot more if it works well.
 */
#if defined (_PTHREAD_USE_ASM_) && defined (_PTHREAD_USE_INLINE)
# pragma inline pthread_getspecific
static void *pthread_getspecific (pthread_key_t key)
{
# ifdef _PTHREAD_ENV_UNIX
    pthread_t		self = pthread_self ();
    return (void *)(key >= self->_Pfield(tsd_count)
		? NULL : self->_Pfield(tsd_values)[key]);
# else
    pthreadTeb_p	self = pthreadTeb ();
    return (void *)(key >= self->_Pfield(tsd_count)
		? NULL : self->_Pfield(tsd_values)[key]);
# endif
}
#else
# ifdef _PTHREAD_USE_MANGLED_NAMES_
#  ifdef _PTHREAD_USE_PTDNAM_
#   pragma extern_prefix "__"
#  else
#   define pthread_getspecific __pthread_getspecific
#  endif
# endif
extern _PTHREAD_CALL_ void *pthread_getspecific (
	pthread_key_t		key);
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix ""
# endif
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_setspecific __pthread_setspecific
# endif
#endif
extern int _PTHREAD_CALL_ pthread_setspecific (
	pthread_key_t		key,
	/* const */ void		*value);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#if (!defined (_PTHREAD_ENV_CXX) && defined (_PTHREAD_ENV_UNIX)) || defined (_PTHREAD_ENV_WIN32)
# define pthread_cleanup_push(_routine_,_arg_) \
    { \
    void (*_XXX_proc) (void *) = (void (*)(void *))(_routine_); \
    void *_XXX_arg = (void *)(_arg_); \
    int _XXX_completed = 0; \
    try {

# define pthread_cleanup_pop(_execute_)	\
    _XXX_completed = 1;} \
    finally { \
	if ((! _XXX_completed) || (_execute_)) _XXX_proc (_XXX_arg);} \
    }
#else
# define pthread_cleanup_push(_routine_,_arg_)	\
    { \
    void (*_XXX_proc) (void *) = (void (*)(void *))(_routine_); \
    void *_XXX_arg = (void *)(_arg_); \
    int _XXX_completed = 0; \
    PTHREAD_TRY_NP {

# define pthread_cleanup_pop(_execute_)	\
    _XXX_completed = 1;} \
    PTHREAD_FINALLY_NP { \
	if ((! _XXX_completed) || (_execute_)) _XXX_proc (_XXX_arg);} \
    PTHREAD_ENDTRY_NP}
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cancel __pthread_cancel
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cancel (
	pthread_t		thread);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

extern int _PTHREAD_CALL_ pthread_setcancelstate (
	int			state,
	int			*oldstate);

extern int _PTHREAD_CALL_ pthread_setcanceltype (
	int			type,
	int			*oldtype);

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_testcancel __pthread_testcancel
# endif
#endif
extern void _PTHREAD_CALL_ pthread_testcancel (void);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_sig_preempt_int_np __pthread_cond_sig_preempt_int_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_sig_preempt_int_np (
	pthread_cond_t	*cond);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_cond_signal_int_np __pthread_cond_signal_int_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_cond_signal_int_np (
	pthread_cond_t	*cond);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_get_expiration_np __pthread_get_expiration_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_get_expiration_np (
	const struct timespec	*delta,
	struct timespec		*abstime);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_delay_np __pthread_delay_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_delay_np (
	const struct timespec	*interval);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_lock_global_np __pthread_lock_global_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_lock_global_np (void);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_unlock_global_np __pthread_unlock_global_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_unlock_global_np (void);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_attr_setguardsize_np __pthread_attr_setguardsize_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_attr_setguardsize_np (
	pthread_attr_t		*attr,
	size_t			guardsize);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

#ifdef _PTHREAD_USE_MANGLED_NAMES_
# ifdef _PTHREAD_USE_PTDNAM_
#  pragma extern_prefix "__"
# else
#  define pthread_attr_getguardsize_np __pthread_attr_getguardsize_np
# endif
#endif
extern int _PTHREAD_CALL_ pthread_attr_getguardsize_np (
	const pthread_attr_t	*attr,
	size_t			*guardsize);
#ifdef _PTHREAD_USE_PTDNAM_
# pragma extern_prefix ""
#endif

extern unsigned long _PTHREAD_CALL_ pthread_getsequence_np (
	pthread_t	thread);

extern unsigned long _PTHREAD_CALL_ pthread_getselfseq_np (void);

/*
 * By default (unless #undef-ed) use an inlined macro version of
 * pthread_getsequence_np, since it's trivial.
 */
#ifdef _PTHREAD_ENV_UNIX
# define pthread_getsequence_np(_t_) (((pthread_t)(_t_))->_Pfield(sequence))
# define pthread_getselfseq_np() (pthread_self ()->_Pfield(sequence))
#else
# define pthread_getsequence_np(_t_) (((pthreadTeb_p)(_t_))->_Pfield(sequence))
# define pthread_getselfseq_np() (pthreadTeb ()->_Pfield(sequence))
#endif

/*
 * The POSIX 1003.1c pthread_getspecific() function doesn't report any
 * errors. If the key value is illegal, or if the thread has no value for the
 * key, it returns NULL without any error indication. Normally that's not a
 * problem. However if you really want to know whether a key is valid, you
 * can use this function. It will return EINVAL if the key is invalid, ENOENT
 * if the key exists but the thread has no value for the key (or value is
 * NULL), and 0 if the key is valid and the thread has a value.
 */
extern int _PTHREAD_CALL_ pthread_key_validate_np (
	pthread_key_t	key);

/*
 * Register desired processor affinity for bound (system scope) threads.
 * This attribute is ignored for threads in process scope, which is the
 * default.
 *
 * These routines are UNSUPPORTED and will be removed in a future release.
 */
#ifdef _PTHREAD_ENV_UNIX
extern int _PTHREAD_CALL_ pthread_attr_setaffinity_np (
	pthread_attr_t		*attr,
	unsigned long		mask);

extern int _PTHREAD_CALL_ pthread_attr_getaffinity_np (
	const pthread_attr_t	*attr,
	unsigned long		*mask);
#endif

/*
 * DECthreads provides three distinct types of mutex. The default (normal or
 * "fast" mutex), which can lock and unlock with completely inline code on
 * Alpha processors, but does no consistency checking; the "recursive" mutex
 * which allows a given thread to lock a mutex it already owns without
 * deadlock (it remains locked until all nested locks are released); and the
 * "errorcheck" mutex, which behaves much like a normal mutex but checks for
 * self deadlock and for releasing a mutex that is either unowned or is owned
 * by another thread. The recursive and errorcheck mutexes require a call
 * into the DECthreads library for every lock and unlock.
 *
 * When synchronization metering is enabled, all mutex types monitor every
 * lock and unlock. A mutex statically initialized using
 * PTHREAD_MUTEX_INITIALIZER is "normal", but is set up so that the first use
 * of the mutex will result in a call into DECthreads (as if the mutex was
 * already locked). If synchronization metering is not enabled, DECthreads
 * will disable this extra state so that future references (in the absence of
 * normal mutex contention) will not cause unnecessary calls. Should you wish
 * to avoid this initial call, you can compile with _PTHREAD_NOMETER_STATIC;
 * normal mutexes statically initialized using this option are not metered.
 */
extern int _PTHREAD_CALL_ pthread_mutexattr_settype_np (
	pthread_mutexattr_t		*attr,
	int				type);

extern int _PTHREAD_CALL_ pthread_mutexattr_gettype_np (
	const pthread_mutexattr_t	*attr,
	int				*type);

extern int _PTHREAD_CALL_ pthread_yield_np (void);

/*
 * Debugging functions
 */
extern void _PTHREAD_CALL_ pthread_debug (void);

/*
 * The pthread_debug_cmd() function returns the status of the last specified
 * operation in the command string.
 */
typedef enum pthreadDbgStatus_t {
    PTHREAD_DBG_SUCCESS = 0,		/* Command was successful */
    PTHREAD_DBG_QUIT = 1,		/* Last command was 'quit' or 'exit' */
    PTHREAD_DBG_NONESEL = 2,		/* No objects selected (e.g., "thread -br") */
    PTHREAD_DBG_SUCCESSPEND = 3,	/* Alternate success */
    PTHREAD_DBG_NOPRIV = -1,		/* No privilege for command */
    PTHREAD_DBG_INVPARAM = -2,		/* Invalid parameter on command */
    PTHREAD_DBG_INVSEQ = -3,		/* Invalid object sequence number given */
    PTHREAD_DBG_INCONSTATE = -4,	/* Inconsistent state for operation */
    PTHREAD_DBG_CORRUPT = -5,		/* Unable to complete; internal corruption */
    PTHREAD_DBG_INVOPTION = -6,		/* Invalid command options */
    PTHREAD_DBG_NOARG = -7,		/* Missing command argument */
    PTHREAD_DBG_INVADDR = -8,		/* Invalid address */
    PTHREAD_DBG_INVCMD = -9,		/* Invalid command */
    PTHREAD_DBG_NULLCMD = -10,		/* No command given */
    PTHREAD_DBG_CONFLICT = -11,		/* Conflicting options */
    PTHREAD_DBG_UNIMPL = -12		/* Unimplemented feature */
    } pthreadDbgStatus_t;

/*
 * Pass a command string to DECthreads debug for interpretation. The
 * argument is a "char *" command string.
 */
extern pthreadDbgStatus_t _PTHREAD_CALL_ pthread_debug_cmd (char *cmd);

#if defined (_PTHREAD_ENV_DECC) && defined (_PTHREAD_ENV_VMS)
# if __DECC_VER >= 50200000
#  pragma __environment __restore
# else
#  pragma __extern_model __restore
# endif
#elif defined (_PTHREAD_ENV_VAXC)
# pragma standard
#endif

#ifdef __cplusplus
    }
#endif

#endif					/* _PTHREAD_USE_D4 */

#endif
